import torch
from torchdt import DType

class DTOptimizer(torch.optim.Optimizer):

    def __init__(self, dtype, device, params, defaults):
        if not issubclass(dtype, DType):
            raise ValueError("dtype must be a subclass of DType.")
        self.dtype = dtype
        self.device = device
        super().__init__(params, defaults)

    def step(self, closure=None):
        raise NotImplementedError("This method should be implemented by subclasses.")

    def zero_grad(self, set_to_none: bool = True): # support set_to_none later
        for group in self.param_groups:
            for param in group['params']:
                param._grad_accum_hook.value.fill_(0.0)

    def convert_params(self, *param_names):
        for group in self.param_groups:
            for name in param_names:
                if name in group and not isinstance(group[name], self.dtype):
                    group[name] = self.dtype(group[name], device=self.device)

    def validate_param(self, param_name, condition):
        for group in self.param_groups:
            if param_name not in group:
                continue
            if not condition(group[param_name]):
                str_val = group[param_name].item() if group[param_name].numel() == 1 else group[param_name]
                raise ValueError(f"Invalid {param_name}: {str_val}")


from .optimizer import DTOptimizer
from .sgd import SGD, TritonSGD
from .adam import Adam, TritonAdam
from .madam import Madam, TritonMadam
from . import lr_scheduler

__all__ = [
    "DTOptimizer",
    "SGD",
    "TritonSGD",
    "Adam",
    "TritonAdam",
    "Madam",
    "TritonMadam",
    "lr_scheduler",
]


import torch
from typing import Any, Dict, Tuple, Type, Callable, Union, Optional

try:
    import torchvision.transforms.v2
    _TV_AVAILABLE = True
except ImportError:
    _TV_AVAILABLE = False

__all__ = [
    "register_collate_dtype_fn",
    "ToDType",
    "DTypeNormalize",
]

def register_collate_dtype_fn(dtype_cls):

    def collate_fn(
            batch,
            *,
            collate_fn_map: Optional[Dict[Union[Type, Tuple[Type, ...]], Callable]] = None,
    ):
        return torch.stack(batch, 0)
    torch.utils.data._utils.collate.default_collate_fn_map.update({dtype_cls: collate_fn})

class ToDType:
    """
    Convert an image-like object (PIL image or numpy.ndarray) with shape
    (H x W x C) in the range [0, 255] to a DType with shape (C x H x W)
    in the range [0., 1.]. This is analogous to torchvision.transforms.ToTensor
    but wraps the output in the given DType.

    Non-floating point inputs are not converted to the DType unless specified
    by the `wrap_all` parameter.
    """

    def __init__(
            self,
            dtype: type,
            wrap_all: bool = False,
            device: Optional[Union[str, torch.device]] = None
    ):
        if not _TV_AVAILABLE:
            raise ImportError(
                "torchvision is required for ToDType transform. "
                "Install it with:  pip install torchvision."
            )

        self.dtype = dtype
        self.wrap_all = wrap_all
        self.device = torch.device(device) if device is not None else None

        self.pipeline = torchvision.transforms.v2.Compose((
            torchvision.transforms.v2.ToImage(),
            torchvision.transforms.v2.ToDtype(dtype.float_dtype, scale=True),
        ))

    def __call__(self, img: Any):
        if torch.is_tensor(img):
            tensor = img
        else:
            tensor = self.pipeline(img)

        if not self.wrap_all and not tensor.is_floating_point():
            # Return unwrapped tensor for integer types unless wrap_all is True
            if self.device is not None:
                tensor = tensor.to(self.device)
            return tensor

        if self.device is not None:
            tensor = tensor.to(self.device)

        return self.dtype(tensor)

class DTypeNormalize:
    """
    Normalize a DType with mean and standard deviation. This is analogous to
    torchvision.transforms.Normalize but performs normalization on the DType.
    """

    def __init__(
        self,
        dtype: type,
        mean: Union[float, Tuple[float, ...]],
        std: Union[float, Tuple[float, ...]],
        device: Optional[Union[str, torch.device]] = None
    ):
        if not isinstance(mean, (float, tuple)):
            raise ValueError(f"mean must be a float or tuple of floats, got {type(mean)}.")
        if not isinstance(std, (float, tuple)):
            raise ValueError(f"std must be a float or tuple of floats, got {type(std)}.")

        self.mean = dtype(mean, device=device)
        self.std = dtype(std, device=device)

        if not (self.mean.ndim == 0 or self.mean.ndim == 1):
            raise ValueError(f"mean must be a scalar or 1D tensor, got shape {self.mean.shape}.")
        if not (self.std.ndim == 0 or self.std.ndim == 1):
            raise ValueError(f"std must be a scalar or 1D tensor, got shape {self.std.shape}.")

        self.mean = self.mean.unsqueeze(-1).unsqueeze(-1)
        self.std = self.std.unsqueeze(-1).unsqueeze(-1)

    def __call__(self, tensor):
        if not tensor.ndim >= 3:
            raise ValueError(f"Input tensor must have at least 3 dimensions (C, H, W), got shape {tensor.shape}.")

        if not (self.mean.numel() == tensor.shape[-3] or self.mean.numel() == 1):
            raise ValueError(
                f"Mean tensor length {self.mean.numel()} does not match "
                f"the number of channels {tensor.shape[-3]}."
            )

        if not (self.std.numel() == tensor.shape[-3] or self.std.numel() == 1):
            raise ValueError(
                f"Std tensor length {self.std.numel()} does not match "
                f"the number of channels {tensor.shape[-3]}."
            )

        return (tensor - self.mean) / self.std


import torch
from torchdt.autograd import DTFunction
from torchdt.ops import register_base_op

@register_base_op("relu")
def dt_relu(ops, x):
    return torch.where(
        ops.lt(x, ops.scalar_from_float(0.0)),
        ops.scalar_from_float(0.0), x
    )

class DTReLUFunction(DTFunction):

    @staticmethod
    def forward(ops, x):
        return ops.relu(x)

    @staticmethod
    def setup_context(ctx, ops, inputs, output):
        ctx.save_for_backward(output)

    @staticmethod
    def backward(ctx, ops, grad_output):
        output, = ctx.saved_tensors
        grad_x = torch.where(ops.eq(output, ops.scalar_from_float(0.0)), ops.scalar_from_float(0.0), grad_output)
        return grad_x

@register_base_op("leaky_relu")
def dt_leaky_relu(ops, x, negative_slope):
    return torch.where(
        ops.lt(x, ops.scalar_from_float(0.0)),
        ops.mul(x, negative_slope), x
    )

class DTLeakyReLUFunction(DTFunction):

    @staticmethod
    def forward(ops, x, negative_slope):
        return ops.leaky_relu(x, negative_slope)

    @staticmethod
    def setup_context(ctx, ops, inputs, output):
        x, negative_slope = inputs
        ctx.save_for_backward(x, negative_slope)

    @staticmethod
    def backward(ctx, ops, grad_output):
        x, negative_slope = ctx.saved_tensors
        grad_x = torch.where(ops.lt(x, ops.scalar_from_float(0.0)), ops.mul(grad_output, negative_slope), grad_output)
        return grad_x, None

@register_base_op("threshold")
def dt_threshold(ops, x, threshold, value):
    return torch.where(
        ops.gt(x, threshold),
        x, value
    )

class DTThresholdFunction(DTFunction):

    @staticmethod
    def forward(ops, x, threshold, value):
        return ops.threshold(x, threshold, value)

    @staticmethod
    def setup_context(ctx, ops, inputs, output):
        x, threshold, _ = inputs
        ctx.save_for_backward(x, threshold)

    @staticmethod
    def backward(ctx, ops, grad_output):
        x, threshold = ctx.saved_tensors
        grad_x = torch.where(ops.gt(x, threshold), grad_output, ops.scalar_from_float(0.0))
        return grad_x, None, None

@register_base_op("tanh")
def dt_tanh(ops, x):
    exp_x = ops.exp(x)
    exp_neg_x = ops.exp(ops.neg(x))
    return ops.div(ops.sub(exp_x, exp_neg_x), ops.add(exp_x, exp_neg_x))

class DTTanhFunction(DTFunction):

    @staticmethod
    def forward(ops, x):
        return ops.tanh(x)

    @staticmethod
    def setup_context(ctx, ops, inputs, output):
        ctx.save_for_backward(output)

    @staticmethod
    def backward(ctx, ops, grad_output):
        output, = ctx.saved_tensors
        grad_x = ops.mul(grad_output, ops.sub(ops.scalar_from_float(1.0), ops.square(output)))
        return grad_x

@register_base_op("sigmoid")
def dt_sigmoid(ops, x):
    exp_neg_x = ops.exp(ops.neg(x))
    return ops.div(ops.scalar_from_float(1.0), ops.add(ops.scalar_from_float(1.0), exp_neg_x))

class DTSigmoidFunction(DTFunction):

    @staticmethod
    def forward(ops, x):
        return ops.sigmoid(x)

    @staticmethod
    def setup_context(ctx, ops, inputs, output):
        ctx.save_for_backward(output)

    @staticmethod
    def backward(ctx, ops, grad_output):
        output, = ctx.saved_tensors
        grad_x = ops.mul(grad_output, ops.mul(output, ops.sub(ops.scalar_from_float(1.0), output)))
        return grad_x

@register_base_op("logsigmoid")
def dt_logsigmoid(ops, x):
    return ops.log(ops.sigmoid(x))

class DTLogSigmoidFunction(DTFunction):

    @staticmethod
    def forward(ops, x):
        return ops.logsigmoid(x)

    @staticmethod
    def setup_context(ctx, ops, inputs, output):
        x, = inputs
        ctx.save_for_backward(x)

    @staticmethod
    def backward(ctx, ops, grad_output):
        x, = ctx.saved_tensors
        grad_x = ops.mul(grad_output, ops.sub(ops.scalar_from_float(1.0), ops.sigmoid(x)))
        return grad_x

@register_base_op("softmin")
def dt_softmin(ops, x, dim=-1):
    exp_neg_x = ops.exp(ops.neg(x))
    sum_exp_neg_x = ops.sum(exp_neg_x, dim=dim, keepdim=True)
    return ops.div(exp_neg_x, sum_exp_neg_x)

class DTSoftminFunction(DTFunction):

    @staticmethod
    def forward(ops, x, dim=None):
        return ops.softmin(x, dim=dim)

    @staticmethod
    def setup_context(ctx, ops, inputs, output):
        _, dim = inputs
        ctx.save_for_backward(output)
        ctx.dim = dim

    @staticmethod
    def backward(ctx, ops, grad_output):
        output, = ctx.saved_tensors
        dot_product = ops.sum(ops.mul(grad_output, output), dim=ctx.dim, keepdim=True)
        grad_x = ops.mul(output, ops.sub(grad_output, dot_product))
        return grad_x, None

@register_base_op("softmax")
def dt_softmax(ops, x, dim=None):
    exp_x = ops.exp(x)
    sum_exp_x = ops.sum(exp_x, dim=dim, keepdim=True)
    return ops.div(exp_x, sum_exp_x)

class DTSoftmaxFunction(DTFunction):

    @staticmethod
    def forward(ops, x, dim=None):
        return ops.softmax(x, dim=dim)

    @staticmethod
    def setup_context(ctx, ops, inputs, output):
        _, dim = inputs
        ctx.save_for_backward(output)
        ctx.dim = dim

    @staticmethod
    def backward(ctx, ops, grad_output):
        output, = ctx.saved_tensors
        dot_product = ops.sum(ops.mul(grad_output, output), dim=ctx.dim, keepdim=True)
        grad_x = ops.mul(output, ops.sub(grad_output, dot_product))
        return grad_x, None

@register_base_op("log_softmax")
def dt_log_softmax(ops, x, dim=None):
    if dim is None:
        m = ops.max(x)
    else:
        m = ops.max(x, dim=dim, keepdim=True)[0] # discard indices

    # subtract the max to prevent overflow (logsumexp trick)
    x_sub_m = ops.sub(x, m)
    log_sum_exp_x_sub_m = ops.log(ops.sum(ops.exp(x_sub_m), dim=dim, keepdim=True))
    return ops.sub(x_sub_m, log_sum_exp_x_sub_m)

class DTLogSoftmaxFunction(DTFunction):

    @staticmethod
    def forward(ops, x, dim=None):
        return ops.log_softmax(x, dim=dim)

    @staticmethod
    def setup_context(ctx, ops, inputs, output):
        _, dim = inputs
        ctx.save_for_backward(output)
        ctx.dim = dim

    @staticmethod
    def backward(ctx, ops, grad_output):
        output, = ctx.saved_tensors
        sum_grad = ops.sum(grad_output, dim=ctx.dim, keepdim=True)
        grad_x = ops.sub(grad_output, ops.mul(ops.exp(output), sum_grad))
        return grad_x, None

@register_base_op("hardtanh")
def dt_hardtanh(ops, x, min_val=-1.0, max_val=1.0):
    result = torch.where(ops.lt(x, min_val), min_val, x)
    result = torch.where(ops.gt(result, max_val), max_val, result)
    return result

class DTHardtanhFunction(DTFunction):

    @staticmethod
    def forward(ops, x, min_val, max_val):
        return ops.hardtanh(x, min_val, max_val)

    @staticmethod
    def setup_context(ctx, ops, inputs, output):
        x, min_val, max_val = inputs
        ctx.save_for_backward(x, min_val, max_val)

    @staticmethod
    def backward(ctx, ops, grad_output):
        x, min_val, max_val = ctx.saved_tensors
        grad_x = torch.where(
            ops.le(x, min_val) | ops.ge(x, max_val),
            ops.scalar_from_float(0.0), grad_output)
        return grad_x, None, None

@register_base_op("glu")
def dt_glu(ops, x, dim=-1):
    half_size = x.size(dim) // 2
    a = x.narrow(dim, 0, half_size)
    b = x.narrow(dim, half_size, half_size)
    return ops.mul(a, ops.sigmoid(b))

class DTGluFunction(DTFunction):

    @staticmethod
    def forward(ops, x, dim=-1):
        return ops.glu(x, dim=dim)

    @staticmethod
    def setup_context(ctx, ops, inputs, output):
        x, dim = inputs
        ctx.save_for_backward(x)
        ctx.dim = dim

    @staticmethod
    def backward(ctx, ops, grad_output):
        x, = ctx.saved_tensors
        half_size = x.size(ctx.dim) // 2
        a = x.narrow(ctx.dim, 0, half_size)
        b = x.narrow(ctx.dim, half_size, half_size)

        sigmoid_b = ops.sigmoid(b)
        grad_a = ops.mul(grad_output, sigmoid_b)
        grad_b = ops.sub(ops.scalar_from_float(1.0), sigmoid_b)
        grad_b = ops.mul(grad_output, ops.mul(a, ops.mul(sigmoid_b, grad_b)))

        grad_x = torch.cat([grad_a, grad_b], dim=ctx.dim)
        return grad_x, None


from . import arithmetic_funcs
from . import comparison_funcs
from . import unary_funcs
from . import activation_funcs
from . import misc_funcs
from . import loss_funcs
from . import layer_funcs
from . import creation_funcs
from . import init_funcs


import torch
from torchdt.optim import DTOptimizer

class Madam(DTOptimizer):

    def __init__(
            self,
            dtype,
            device,
            params,
            lr=0.01,
            beta=0.999,
            eps=1e-8,
            p_scale=3.0,
            g_bound=10.0,
            use_pow=False,
            *,
            maximize=False,
    ):
        defaults = dict(
            lr=lr,
            beta=beta,
            eps=eps,
            p_scale=p_scale,
            g_bound=g_bound,
            use_pow=use_pow,
            maximize=maximize,
        )
        super().__init__(dtype, device, params, defaults)
        self.convert_params("lr", "beta", "eps", "p_scale", "g_bound")

        self.validate_param("lr", lambda lr: lr >= torch.tensor(0.0, device=device))
        self.validate_param("eps", lambda eps: eps > torch.tensor(0.0, device=device))
        self.validate_param("beta",lambda beta: torch.tensor(0.0, device=device) < beta < torch.tensor(1.0, device=device))
        self.validate_param("p_scale", lambda p_scale: p_scale > torch.tensor(0.0, device=device))
        self.validate_param("g_bound", lambda g_bound: g_bound > torch.tensor(0.0, device=device))

    @torch.no_grad()
    def step(self, closure=None):
        """Performs a single optimization step."""

        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:
            lr = group["lr"]
            beta = group["beta"]
            eps = group["eps"]
            p_scale = group["p_scale"]
            g_bound = group["g_bound"]
            use_pow = group["use_pow"]
            maximize = group["maximize"]

            for p in group["params"]:
                if p.grad is None:
                    continue

                grad = p.grad
                state = self.state[p]

                if len(state) == 0:
                    # First time we see this parameter
                    rms = torch.sqrt(torch.mean(p * p))
                    state["max"] = p_scale * rms
                    state["step"] = self.dtype(0.0, device=self.device)
                    state["exp_avg_sq"] = torch.zeros_like(p, dtype=self.dtype, device=self.device)

                max = state["max"]
                step = state["step"] + torch.tensor(1.0, device=self.device)
                exp_avg_sq = state["exp_avg_sq"]

                exp_avg_sq = (beta * exp_avg_sq) + (grad * grad * (torch.tensor(1.0, device=self.device) - beta))
                corr_exp_avg_sq = exp_avg_sq / (torch.tensor(1.0, device=self.device) - beta ** step) + eps

                g_normed = grad / torch.sqrt(corr_exp_avg_sq)
                g_clipped = torch.clamp(g_normed, -g_bound, g_bound)
                delta = lr * g_clipped * torch.sign(p)

                if not maximize:
                    delta = -delta

                if use_pow:
                    mul_update = p * torch.exp(delta)
                else:
                    mul_update = p * (torch.tensor(1.0, device=self.device) + delta)

                p.data.copy_(torch.clamp(mul_update, -max, max))

                state["step"] = step
                state["exp_avg_sq"] = exp_avg_sq

        return loss


class TritonMadam(DTOptimizer):

    def __init__(
            self,
            dtype,
            device,
            params,
            lr=0.01,
            beta=0.999,
            eps=1e-8,
            p_scale=3.0,
            g_bound=10.0,
            use_pow=False,
            *,
            maximize=False,
    ):
        defaults = dict(
            lr=lr,
            beta=beta,
            eps=eps,
            p_scale=p_scale,
            g_bound=g_bound,
            use_pow=use_pow,
            maximize=maximize,
        )
        super().__init__(dtype, device, params, defaults)
        self.convert_params("lr", "beta", "eps", "p_scale", "g_bound")

        self.validate_param("lr", lambda lr: lr >= torch.tensor(0.0, device=device))
        self.validate_param("eps", lambda eps: eps > torch.tensor(0.0, device=device))
        self.validate_param("beta",lambda beta: torch.tensor(0.0, device=device) < beta < torch.tensor(1.0, device=device))
        self.validate_param("p_scale", lambda p_scale: p_scale > torch.tensor(0.0, device=device))
        self.validate_param("g_bound", lambda g_bound: g_bound > torch.tensor(0.0, device=device))

    @torch.no_grad()
    def step(self, closure=None):
        """Performs a single optimization step."""

        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:
            lr = group["lr"]
            beta = group["beta"]
            eps = group["eps"]
            p_scale = group["p_scale"]
            g_bound = group["g_bound"]
            use_pow = group["use_pow"]
            maximize = group["maximize"]

            for p in group["params"]:
                if p.grad is None:
                    continue

                grad = p.grad
                state = self.state[p]

                if len(state) == 0:
                    rms = torch.sqrt(torch.mean(p * p))
                    state["max"] = p_scale * rms
                    state["step"] = self.dtype(0.0, device=self.device)
                    state["exp_avg_sq"] = torch.zeros_like(p, dtype=self.dtype, device=self.device)._int

                max = state["max"]
                step = state["step"] + torch.tensor(1.0, device=self.device)
                exp_avg_sq = state["exp_avg_sq"]

                bias_corr = torch.tensor(1.0, device=self.device) - beta ** step
                new_exp_avg_sq = self.dtype.ops.triton_madam_step(
                    p._int, grad._int, exp_avg_sq,
                    lr._int, beta._int, eps._int,
                    g_bound._int, max._int, bias_corr._int,
                    use_pow, maximize,
                )

                state["step"] = step
                state["exp_avg_sq"] = new_exp_avg_sq

        return loss


from torch.utils.cpp_extension import BuildExtension, include_paths
from pathlib import Path
import warnings
import os

# The exact exception raised by a compiler/linker failure can vary
# between setuptools versions, so we catch the whole family.
try:
    # setuptools ≥ 68
    from setuptools.errors import (
        CompileError,
        LinkError,
        PlatformError,
    )
except ImportError:
    # setuptools < 68
    from distutils.errors import ( # type: ignore
        CompileError,
        LinkError,
        DistutilsPlatformError as PlatformError,
    )

class BuildCxxExtension(BuildExtension):
    """
    Build all .cpp files in csrc/ and include Torch headers.
    Fails hard if compilation fails.
    """

    user_options = BuildExtension.user_options + [
        ("no-cpp", None, "Do not build the C++ extension"),
    ]
    boolean_options = getattr(BuildExtension, "boolean_options", []) + ["no-cpp"]

    def initialize_options(self):
        super().initialize_options()
        self.no_cpp = False

    def finalize_options(self):
        super().finalize_options()
        if os.environ.get("TORCHDT_NO_CPP", "").lower() in {"1", "true", "yes", "on"}:
            self.no_cpp = True

    def run(self):
        try:
            super().run()
        except (CompileError, LinkError, PlatformError) as exc:
            warnings.warn(
                "C++ backend for xlnstorch was not built; "
                "installation will succeed but will run in pure-python mode.",
                RuntimeWarning,
            )

    def build_extensions(self):
        if self.no_cpp:
            self.extensions = []
            return

        src_dir = Path(__file__).resolve().parent / "csrc"
        include_dir = Path(__file__).resolve().parent / "include"
        cpp_files = [str(p) for p in src_dir.glob("*.cpp")]

        torch_includes = include_paths()

        for ext in self.extensions:
            if ext.name == "torchdt._C":
                ext.sources = cpp_files
            ext.include_dirs = list(ext.include_dirs or [])
            ext.include_dirs += torch_includes
            ext.include_dirs += [str(include_dir)]

        super().build_extensions()



import math
import torch
from torchdt.autograd import DTFunction
from torchdt.ops import register_base_op

@register_base_op("linear")
def dt_linear(ops, x, weight, bias=None):
    output = ops.matmul(x, weight.transpose(-2, -1))
    if bias is not None:
        output = ops.add(output, bias)
    return output

class DTLinearFunction(DTFunction):

    @staticmethod
    def forward(ops, x, weight, bias=None):
        return ops.linear(x, weight, bias)

    @staticmethod
    def setup_context(ctx, ops, inputs, output):
        x, A, bias = inputs
        ctx.biased = bias is not None
        ctx.save_for_backward(x, A)

    @staticmethod
    def backward(ctx, ops, grad_output):
        x, A = ctx.saved_tensors
        grad_x = ops.matmul(grad_output, A)
        out_features = A.shape[0]
        in_features = A.shape[1]

        grad_output_2d = grad_output.reshape(-1, out_features)
        x_2d = x.reshape(-1, in_features)
        grad_output_T = grad_output_2d.transpose(0, 1)
        grad_A = ops.matmul(grad_output_T, x_2d)

        if ctx.biased:
            if grad_output.dim() == 1:
                grad_bias = grad_output
            else:
                grad_bias = ops.sum(grad_output, dim=tuple(range(grad_output.dim() - 1)))
        else:
            grad_bias = None

        return grad_x, grad_A, grad_bias

@register_base_op("dropout")
def dt_dropout(ops, x, p=0.5):
    mask = ops.from_float(torch.bernoulli(torch.full(x.shape, 1 - p)))
    return ops.mul(x, mask)

class DTDropoutFunction(DTFunction):

    @staticmethod
    def forward(ops, x, p=0.5):
        return ops.dropout(x, p)

    @staticmethod
    def setup_context(ctx, ops, inputs, output):
        _, p = inputs
        ctx.save_for_backward(output)
        ctx.p = p

    @staticmethod
    def backward(ctx, ops, grad_output):
        output, = ctx.saved_tensors
        grad_x = torch.where(
            output == ops.scalar_from_float(0.0),
            ops.scalar_from_float(0.0),
            ops.mul(grad_output, ops.scalar_from_float(1 / (1 - ctx.p))))
        return grad_x, None

@register_base_op("conv2d")
def dt_conv2d(ops, x, weight, bias=None, stride=1, padding=0, dilation=1, groups=1):
    # Handle stride, padding, dilation as tuples
    if isinstance(stride, int):
        stride = (stride, stride)
    if isinstance(padding, int):
        padding = (padding, padding)
    if isinstance(dilation, int):
        dilation = (dilation, dilation)
    stride_h, stride_w = stride
    pad_h, pad_w = padding
    dil_h, dil_w = dilation

    # Add batch dimension if needed
    squeeze_batch = False
    if x.dim() == 3:
        squeeze_batch = True
        x = x.unsqueeze(0)

    N, C_in, H_in, W_in = x.shape # [batch, in_channels, height, width]
    C_out, C_w, K_H, K_W = weight.shape  # [out_channels, in_per_group, kh, kw]
    assert C_in % groups == 0
    assert C_out % groups == 0
    assert C_w == C_in // groups
    g_Cin = C_in // groups
    g_Cout = C_out // groups

    # Pad input
    if pad_h > 0 or pad_w > 0:
        x_padded = torch.nn.functional.pad(
            x, (pad_w, pad_w, pad_h, pad_h), value=ops.scalar_from_float(0.0).item()
        )
    else:
        x_padded = x

    H_pad, W_pad = x_padded.shape[2:]
    # Output shape calculation as per PyTorch
    H_out = (H_in + 2 * pad_h - dil_h * (K_H - 1) - 1) // stride_h + 1
    W_out = (W_in + 2 * pad_w - dil_w * (K_W - 1) - 1) // stride_w + 1
    out = ops.zeros(N, C_out, H_out, W_out)

    for n in range(N):
        for g in range(groups):
            inp_group = x_padded[n, g * g_Cin : (g + 1) * g_Cin] # [g_Cin, H_pad, W_pad]
            for c_out in range(g * g_Cout, (g + 1) * g_Cout):
                for h in range(H_out):
                    for w in range(W_out):
                        h_start = h * stride_h
                        w_start = w * stride_w
                        h_end = h_start + K_H * dil_h
                        w_end = w_start + K_W * dil_w
                        # Extract appropriate input window
                        inp_slice = inp_group[:, h_start:h_end:dil_h, w_start:w_end:dil_w] # shape [g_Cin, K_H, K_W]
                        out[n, c_out, h, w] = ops.sum(ops.mul(inp_slice, weight[c_out]))
                        if bias is not None:
                            out[n, c_out, h, w] = ops.add(out[n, c_out, h, w], bias[c_out])

    if squeeze_batch:
        out = out.squeeze(0)
    return out

class DTConv2dFunction(DTFunction):

    @staticmethod
    def forward(ops, x, weight, bias=None, stride=1, padding=0, dilation=1, groups=1):
        return ops.conv2d(x, weight, bias, stride, padding, dilation, groups)

    @staticmethod
    def setup_context(ctx, ops, inputs, output):
        x, weight, bias, stride, padding, dilation, groups = inputs
        ctx.save_for_backward(x, weight, bias)
        ctx.stride = stride
        ctx.padding = padding
        ctx.dilation = dilation
        ctx.groups = groups

    @staticmethod
    def backward(ctx, ops, grad_output):
        x, weight, bias = ctx.saved_tensors

        # Canonicalize params
        stride = ctx.stride if isinstance(ctx.stride, tuple) else (ctx.stride, ctx.stride)
        padding = ctx.padding if isinstance(ctx.padding, tuple) else (ctx.padding, ctx.padding)
        dilation = ctx.dilation if isinstance(ctx.dilation, tuple) else (ctx.dilation, ctx.dilation)
        stride_h, stride_w = stride
        pad_h, pad_w = padding
        dil_h, dil_w = dilation
        groups = ctx.groups

        squeeze_batch = (x.dim() == 3)
        if grad_output.dim() == 3:
            grad_output = grad_output.unsqueeze(0)
            squeeze_batch = True

        N, C_out, H_out, W_out = grad_output.shape
        C_in = weight.shape[1] * groups
        K_H, K_W = weight.shape[2:]
        g_Cin = C_in // groups
        g_Cout = C_out // groups

        # Pad input for correct alignment
        if pad_h > 0 or pad_w > 0:
            x_padded = torch.nn.functional.pad(x, (pad_w, pad_w, pad_h, pad_h), value=ops.scalar_from_float(0.0).item())
        else:
            x_padded = x

        H_pad = x_padded.shape[2]
        W_pad = x_padded.shape[3]

        grad_x_padded = ops.zeros(N, C_in, H_pad, W_pad)
        grad_weight = ops.zeros_like(weight)

        # Input gradient: for each pixel in input, sum all contributions from grad_output via the receptive field
        for n in range(N):
            for g in range(groups):
                in_start = g * g_Cin
                in_end = (g + 1) * g_Cin
                out_start = g * g_Cout
                out_end = (g + 1) * g_Cout
                for c_in in range(g_Cin):
                    for h_in in range(H_pad):
                        for w_in in range(W_pad):
                            grad = ops.scalar_from_float(0.0)
                            for c_out in range(out_start, out_end):
                                w = weight[c_out, c_in, :, :]
                                for k_h in range(K_H):
                                    for k_w in range(K_W):
                                        # Figure out which output position this input contributes to
                                        h_out_nom = h_in - k_h * dil_h
                                        w_out_nom = w_in - k_w * dil_w
                                        if h_out_nom % stride_h == 0 and w_out_nom % stride_w == 0:
                                            h_out = h_out_nom // stride_h
                                            w_out = w_out_nom // stride_w
                                            if (0 <= h_out < H_out) and (0 <= w_out < W_out):
                                                grad = ops.add(
                                                    grad,
                                                    ops.mul(grad_output[n, c_out, h_out, w_out], w[k_h, k_w])
                                                )
                            grad_x_padded[n, in_start + c_in, h_in, w_in] = grad

        # Remove padding to match original input shape
        if pad_h > 0 or pad_w > 0:
            grad_x = grad_x_padded[:, :, pad_h: H_pad - pad_h, pad_w: W_pad - pad_w]
        else:
            grad_x = grad_x_padded

        # Weight gradient: correlate grad_output windows with input
        for g in range(groups):
            in_start = g * g_Cin
            in_end = (g + 1) * g_Cin
            out_start = g * g_Cout
            out_end = (g + 1) * g_Cout
            for c_out in range(out_start, out_end):
                for c_in in range(g_Cin):
                    for k_h in range(K_H):
                        for k_w in range(K_W):
                            grad = ops.scalar_from_float(0.0)
                            for n in range(N):
                                for h_out in range(H_out):
                                    for w_out in range(W_out):
                                        h_in = h_out * stride_h + k_h * dil_h
                                        w_in = w_out * stride_w + k_w * dil_w
                                        inp_padded = x_padded[n, in_start + c_in, :, :]
                                        if (0 <= h_in < inp_padded.size(0)) and (0 <= w_in < inp_padded.size(1)):
                                            grad = ops.add(
                                                grad,
                                                ops.mul(grad_output[n, c_out, h_out, w_out], inp_padded[h_in, w_in])
                                            )
                            grad_weight[c_out, c_in, k_h, k_w] = grad

        # Bias gradient: sum grad_output over batch, spatial dims
        if bias is not None:
            grad_bias = ops.sum(grad_output, dim=[0, 2, 3])
        else:
            grad_bias = None

        if squeeze_batch:
            grad_x = grad_x.squeeze(0)
        return grad_x, grad_weight, grad_bias, None, None, None, None

@register_base_op("avg_pool2d")
def dt_avg_pool2d(ops, x, kernel_size, stride=None, padding=0, ceil_mode=False, count_include_pad=True, divisor_override=None):
    if stride is None:
        stride = kernel_size

    if isinstance(kernel_size, int):
        kernel_size = (kernel_size, kernel_size)
    if isinstance(stride, int):
        stride = (stride, stride)
    if isinstance(padding, int):
        padding = (padding, padding)

    kernel_w, kernel_h = kernel_size
    stride_h, stride_w = stride
    pad_h, pad_w = padding

    if x.dim() == 3:
        x = x.unsqueeze(0)
        squeeze_batch = True
    else:
        squeeze_batch = False

    N, C, H_in, W_in = x.shape
    if pad_h > 0 or pad_w > 0:
        x_padded = torch.nn.functional.pad(x, (pad_w, pad_w, pad_h, pad_h), value=ops.scalar_from_float(0.0).item())
    else:
        x_padded = x

    if ceil_mode:
        H_out = int(math.ceil((H_in + 2 * pad_h - kernel_h) / stride_h)) + 1
        W_out = int(math.ceil((W_in + 2 * pad_w - kernel_w) / stride_w)) + 1
    else:
        H_out = (H_in + 2 * pad_h - kernel_h) // stride_h + 1
        W_out = (W_in + 2 * pad_w - kernel_w) // stride_w + 1

    out = ops.zeros(N, C, H_out, W_out)
    kernel_area_dt = ops.scalar_from_float(kernel_h * kernel_w)

    for n in range(N):
        for c in range(C):
            for h_out in range(H_out):
                for w_out in range(W_out):
                    h_start = h_out * stride_h
                    h_end = h_start + kernel_h
                    w_start = w_out * stride_w
                    w_end = w_start + kernel_w
                    if h_end > x_padded.size(-2) or w_end > x_padded.size(-1):
                        continue

                    window = x_padded[n, c, h_start:h_end, w_start:w_end]
                    sm = ops.sum(window)

                    if divisor_override is not None:
                        divisor = divisor_override
                    elif count_include_pad:
                        divisor = kernel_area_dt
                    else:
                        left_pad = max(0, pad_w - w_start)
                        right_pad = max(0, w_end - (W_in + pad_w))
                        top_pad = max(0, pad_h - h_start)
                        bot_pad = max(0, h_end - (H_in + pad_h))
                        valid_h = kernel_h - (top_pad + bot_pad)
                        valid_w = kernel_w - (left_pad + right_pad)
                        valid_count = max(valid_h, 0) * max(valid_w, 0)
                        divisor = ops.scalar_from_float(max(valid_count, 1))

                    out[n, c, h_out, w_out] = ops.div(sm, divisor)

    if squeeze_batch:
        out = out.squeeze(0)
    return out

class DTAvgPool2dFunction(DTFunction):

    @staticmethod
    def forward(ops, x, kernel_size, stride=None, padding=0, ceil_mode=False, count_include_pad=True, divisor_override=None):
        return ops.avg_pool2d(x, kernel_size, stride, padding, ceil_mode, count_include_pad, divisor_override)

    @staticmethod
    def setup_context(ctx, ops, inputs, output):
        x, kernel_size, stride, padding, ceil_mode, count_include_pad, divisor_override = inputs
        if stride is None:
            stride = kernel_size
        if isinstance(kernel_size, int):
            kernel_size = (kernel_size, kernel_size)
        if isinstance(stride, int):
            stride = (stride, stride)
        if isinstance(padding, int):
            padding = (padding, padding)
        if divisor_override is None:
            ctx.divisor_override = False
            ctx.save_for_backward(x)
        else:
            ctx.divisor_override = True
            ctx.save_for_backward(x, divisor_override)
        ctx.kernel_size = kernel_size
        ctx.stride = stride
        ctx.padding = padding
        ctx.ceil_mode = ceil_mode
        ctx.count_include_pad = count_include_pad

    @staticmethod
    def backward(ctx, ops, grad_output):
        if ctx.divisor_override:
            x, divisor_override = ctx.saved_tensors
        else:
            x, = ctx.saved_tensors

        kernel_h, kernel_w = ctx.kernel_size
        stride_h, stride_w = ctx.stride
        pad_h, pad_w = ctx.padding

        if x.dim() == 3:
            x = x.unsqueeze(0)
            grad_output = grad_output.unsqueeze(0)
            squeeze_batch = True
        else:
            squeeze_batch = False

        N, C, H_in, W_in = x.shape
        H_out, W_out = grad_output.shape[-2:]

        grad_x = ops.zeros_like(x)
        kernel_area_dt = ops.scalar_from_float(kernel_h * kernel_w)

        for n in range(N):
            for c in range(C):
                for h_out in range(H_out):
                    for w_out in range(W_out):
                        h_start = h_out * stride_h
                        h_end = h_start + kernel_h
                        w_start = w_out * stride_w
                        w_end = w_start + kernel_w

                        if ctx.divisor_override:
                            divisor = divisor_override
                        elif ctx.count_include_pad:
                            divisor = kernel_area_dt
                        else:
                            left_pad = max(0, pad_w - w_start)
                            right_pad = max(0, w_end - (W_in + pad_w))
                            top_pad = max(0, pad_h - h_start)
                            bot_pad = max(0, h_end - (H_in + pad_h))
                            valid_h = kernel_h - (top_pad + bot_pad)
                            valid_w = kernel_w - (left_pad + right_pad)
                            valid_count = max(valid_h, 0) * max(valid_w, 0)
                            divisor = ops.scalar_from_float(max(valid_count, 1))

                        grad = grad_output[n, c, h_out, w_out]
                        for i in range(kernel_h):
                            for j in range(kernel_w):
                                h_idx = h_start + i - pad_h
                                w_idx = w_start + j - pad_w
                                if 0 <= h_idx < H_in and 0 <= w_idx < W_in:
                                    grad_x[n, c, h_idx, w_idx] = ops.add(
                                        grad_x[n, c, h_idx, w_idx],
                                        ops.div(grad, divisor)
                                    )

        if squeeze_batch:
            grad_x = grad_x.squeeze(0)
        return grad_x, None, None, None, None, None, None

@register_base_op("adaptive_avg_pool2d")
def dt_adaptive_avg_pool2d(ops, x, output_size):
    if isinstance(output_size, int):
        H_out, W_out = output_size, output_size
    else:
        assert len(output_size) == 2, "output_size must be int or tuple of length 2"
        H_out, W_out = output_size

    if x.dim() == 3:
        x = x.unsqueeze(0)
        squeeze_batch = True
    else:
        squeeze_batch = False

    N, C, H_in, W_in = x.shape
    out = ops.zeros(N, C, H_out, W_out)

    for n in range(N):
        for c in range(C):
            for h_out in range(H_out):
                h_start = int(math.floor(h_out * H_in / H_out))
                h_end = int(math.ceil((h_out + 1) * H_in / H_out))
                for w_out in range(W_out):
                    w_start = int(math.floor(w_out * W_in / W_out))
                    w_end = int(math.ceil((w_out + 1) * W_in / W_out))

                    window = x[n, c, h_start:h_end, w_start:w_end]
                    sm = ops.sum(window)
                    divisor = ops.scalar_from_float(max((h_end - h_start) * (w_end - w_start), 1))
                    out[n, c, h_out, w_out] = ops.div(sm, divisor)

    if squeeze_batch:
        out = out.squeeze(0)
    return out

class DTAdaptiveAvgPool2dFunction(DTFunction):

    @staticmethod
    def forward(ops, x, output_size):
        return ops.adaptive_avg_pool2d(x, output_size)

    @staticmethod
    def setup_context(ctx, ops, inputs, output):
        x, output_size = inputs
        ctx.save_for_backward(x)
        ctx.output_size = output_size

    @staticmethod
    def backward(ctx, ops, grad_output):
        x, = ctx.saved_tensors

        if isinstance(ctx.output_size, int):
            H_out, W_out = ctx.output_size, ctx.output_size
        else:
            H_out, W_out = ctx.output_size

        if x.dim() == 3:
            x = x.unsqueeze(0)
            grad_output = grad_output.unsqueeze(0)
            squeeze_batch = True
        else:
            squeeze_batch = False

        N, C, H_in, W_in = x.shape
        grad_x = ops.zeros_like(x)

        for n in range(N):
            for c in range(C):
                for h_out in range(H_out):
                    h_start = int(math.floor(h_out * H_in / H_out))
                    h_end = int(math.ceil((h_out + 1) * H_in / H_out))
                    for w_out in range(W_out):
                        w_start = int(math.floor(w_out * W_in / W_out))
                        w_end = int(math.ceil((w_out + 1) * W_in / W_out))

                        divisor = ops.scalar_from_float(max((h_end - h_start) * (w_end - w_start), 1))
                        grad = ops.div(grad_output[n, c, h_out, w_out], divisor)

                        for i in range(h_start, h_end):
                            for j in range(w_start, w_end):
                                grad_x[n, c, i, j] = ops.add(grad_x[n, c, i, j], grad)

        if squeeze_batch:
            grad_x = grad_x.squeeze(0)
        return grad_x, None

@register_base_op("max_pool2d")
def dt_max_pool2d(ops, x, kernel_size, stride=None, padding=0, dilation=1, ceil_mode=False, return_indices=False):
    if stride is None:
        stride = kernel_size

    if isinstance(kernel_size, int):
        kernel_size = (kernel_size, kernel_size)
    if isinstance(stride, int):
        stride = (stride, stride)
    if isinstance(padding, int):
        padding = (padding, padding)
    if isinstance(dilation, int):
        dilation = (dilation, dilation)

    kernel_h, kernel_w = kernel_size
    stride_h, stride_w = stride
    pad_h, pad_w = padding
    dil_h, dil_w = dilation

    if x.dim() == 3:
        x = x.unsqueeze(0)
        squeeze_batch = True
    else:
        squeeze_batch = False

    N, C, H_in, W_in = x.shape
    if pad_h > 0 or pad_w > 0:
        x_padded = torch.nn.functional.pad(x, (pad_w, pad_w, pad_h, pad_h),
                                           value=ops.scalar_from_float(float("-inf")).item())
    else:
        x_padded = x

    eff_kh = (kernel_h - 1) * dil_h + 1
    eff_kw = (kernel_w - 1) * dil_w + 1
    if ceil_mode:
        H_out = int(math.ceil((H_in + 2 * pad_h - eff_kh) / stride_h)) + 1
        W_out = int(math.ceil((W_in + 2 * pad_w - eff_kw) / stride_w)) + 1
    else:
        H_out = (H_in + 2 * pad_h - eff_kh) // stride_h + 1
        W_out = (W_in + 2 * pad_w - eff_kw) // stride_w + 1

    out_vals = ops.zeros(N, C, H_out, W_out)
    if return_indices:
        out_idx = torch.empty((N, C, H_out, W_out), dtype=torch.int64, device=x.device)

    padded_H = x_padded.size(-2)
    padded_W = x_padded.size(-1)

    for n in range(N):
        for c in range(C):
            for h_out in range(H_out):
                h_start = h_out * stride_h
                h_end = h_start + eff_kh

                if h_start >= padded_H:
                    continue

                for w_out in range(W_out):
                    w_start = w_out * stride_w
                    w_end = w_start + eff_kw

                    if w_start >= padded_W:
                        continue

                    h_end_eff = min(h_end, padded_H)
                    w_end_eff = min(w_end, padded_W)

                    window = x_padded[n, c, h_start:h_end_eff:dil_h, w_start:w_end_eff:dil_w]
                    window_flat = window.reshape(-1)

                    # pass dim to get indices
                    mx_val, mx_idx = ops.max(window_flat, dim=0)
                    out_vals[n, c, h_out, w_out] = mx_val

                    len_w = (w_end_eff - w_start + dil_w - 1) // dil_w

                    if return_indices:
                        mx_idx_int = int(mx_idx)
                        i = mx_idx_int // len_w
                        j = mx_idx_int % len_w
                        h_in_idx = (h_start + i * dil_h) - pad_h
                        w_in_idx = (w_start + j * dil_w) - pad_w
                        out_idx[n, c, h_out, w_out] = h_in_idx * W_in + w_in_idx

    if squeeze_batch:
        out_vals = out_vals.squeeze(0)
        if return_indices:
            out_idx = out_idx.squeeze(0)

    if return_indices:
        return out_vals, out_idx
    else:
        return out_vals

class DTMaxPool2dFunction(DTFunction):

    output_indices = [0]

    @staticmethod
    def forward(ops, x, kernel_size, stride=None, padding=0, dilation=1, ceil_mode=False, return_indices=False):
        return ops.max_pool2d(x, kernel_size, stride, padding, dilation, ceil_mode, return_indices)

    @staticmethod
    def setup_context(ctx, ops, inputs, output):
        x, kernel_size, stride, padding, dilation, ceil_mode, return_indices = inputs

        if stride is None:
            stride = kernel_size

        if isinstance(kernel_size, int):
            kernel_size = (kernel_size, kernel_size)
        if isinstance(stride, int):
            stride = (stride, stride)
        if isinstance(padding, int):
            padding = (padding, padding)
        if isinstance(dilation, int):
            dilation = (dilation, dilation)

        ctx.save_for_backward(x)
        ctx.kernel_size = kernel_size
        ctx.stride = stride
        ctx.padding = padding
        ctx.dilation = dilation
        ctx.ceil_mode = ceil_mode
        ctx.return_indices = return_indices

    @staticmethod
    def backward(ctx, ops, grad_output, grad_out_indices=None):
        x, = ctx.saved_tensors

        kernel_h, kernel_w = ctx.kernel_size
        stride_h, stride_w = ctx.stride
        pad_h, pad_w = ctx.padding
        dil_h, dil_w = ctx.dilation

        if x.dim() == 3:
            x = x.unsqueeze(0)
            grad_output = grad_output.unsqueeze(0)
            squeeze_batch = True
        else:
            squeeze_batch = False

        N, C, H_in, W_in = x.shape

        if pad_h > 0 or pad_w > 0:
            x_padded = torch.nn.functional.pad(x, (pad_w, pad_w, pad_h, pad_h),
                                               value=ops.scalar_from_float(float("-inf")).item())
        else:
            x_padded = x

        padded_H, padded_W = x_padded.shape[-2:]
        eff_kh = (kernel_h - 1) * dil_h + 1
        eff_kw = (kernel_w - 1) * dil_w + 1

        if ctx.ceil_mode:
            H_out = int(math.ceil((H_in + 2 * pad_h - eff_kh) / stride_h)) + 1
            W_out = int(math.ceil((W_in + 2 * pad_w - eff_kw) / stride_w)) + 1
        else:
            H_out = (H_in + 2 * pad_h - eff_kh) // stride_h + 1
            W_out = (W_in + 2 * pad_w - eff_kw) // stride_w + 1

        grad_padded = ops.zeros_like(x_padded)
        for n in range(N):
            for c in range(C):
                for h_out in range(H_out):
                    h_start = h_out * stride_h
                    h_end = h_start + eff_kh

                    if h_start >= padded_H:
                        continue

                    for w_out in range(W_out):
                        w_start = w_out * stride_w
                        w_end = w_start + eff_kw

                        if w_start >= padded_W:
                            continue

                        h_end_eff = min(h_end, padded_H)
                        w_end_eff = min(w_end, padded_W)

                        len_h = (h_end_eff - h_start + dil_h - 1) // dil_h
                        len_w = (w_end_eff - w_start + dil_w - 1) // dil_w
                        if len_h <= 0 or len_w <= 0:
                            continue

                        window = x_padded[n, c, h_start:h_end_eff:dil_h, w_start:w_end_eff:dil_w]
                        window_flat = window.reshape(-1)

                        _, mx_idx = ops.max(window_flat, dim=0)

                        mx_idx = int(mx_idx)
                        i = mx_idx // len_w
                        j = mx_idx %  len_w

                        h_in_idx = h_start + i * dil_h
                        w_in_idx = w_start + j * dil_w

                        grad_padded[n, c, h_in_idx, w_in_idx] = ops.add(grad_padded[n, c, h_in_idx, w_in_idx],
                                                                        grad_output[n, c, h_out, w_out])

        if pad_h > 0 or pad_w > 0:
            grad_padded = grad_padded[:, :, pad_h:pad_h + H_in, pad_w:pad_w + W_in]
        grad_x = grad_padded.squeeze(0) if squeeze_batch else grad_padded
        return grad_x, None, None, None, None, None, None

@register_base_op("batch_norm")
def dt_batch_norm(ops, x, running_mean, running_var, momentum, eps, weight=None, bias=None, training=True):
    red_dims = tuple(i for i in range(x.dim()) if i != 1)

    if training:
        batch_mean = ops.mean(x, dim=red_dims, keepdim=True)
        batch_var = ops.var(x, ops.scalar_from_float(0.0), dim=red_dims, keepdim=True)
        batch_var_corrected = ops.var(x, ops.scalar_from_float(1.0), dim=red_dims, keepdim=True)

        with torch.no_grad():
            one_minus_momentum = ops.sub(ops.scalar_from_float(1.0), momentum)
            new_running_mean = ops.add(
                ops.mul(one_minus_momentum, running_mean),
                ops.mul(momentum, batch_mean.squeeze()))
            running_mean.copy_(new_running_mean)
            new_running_var = ops.add(
                ops.mul(one_minus_momentum, running_var),
                ops.mul(momentum, batch_var_corrected.squeeze()))
            running_var.copy_(new_running_var)

        mean = batch_mean
        var = batch_var

    else:
        mean = running_mean.view(1, -1, *([1] * (x.dim() - 2)))
        var = running_var.view(1, -1, *([1] * (x.dim() - 2)))

    var_eps = ops.add(var, eps)
    inv_std = ops.div(ops.scalar_from_float(1.0), ops.sqrt(var_eps))
    x_centered = ops.sub(x, mean)
    x_hat = ops.mul(x_centered, inv_std)

    if weight is not None:
        y = ops.mul(x_hat, weight.view(1, -1, *([1] * (x.dim() - 2))))
    else:
        y = x_hat

    if bias is not None:
        y = ops.add(y, bias.view(1, -1, *([1] * (x.dim() - 2))))
    return y

class DTBatchNormFunction(DTFunction):

    @staticmethod
    def forward(ops, x, running_mean, running_var, weight=None, bias=None, training=False, momentum=0.1, eps=1e-5):
        return ops.batch_norm(x, running_mean, running_var, momentum, eps, weight, bias, training)

    @staticmethod
    def setup_context(ctx, ops, inputs, output):
        x, running_mean, running_var, weight, bias, training, _, eps = inputs
        ctx.red_dims = tuple(i for i in range(x.dim()) if i != 1)
        ctx.training = training

        if training:
            mean = ops.mean(x, dim=ctx.red_dims, keepdim=True)
            var = ops.var(x, ops.scalar_from_float(1.0), dim=ctx.red_dims, keepdim=True)
        else:
            mean = running_mean.reshape(1, -1, *([1] * (x.dim() - 2)))
            var = running_var.reshape(1, -1, *([1] * (x.dim() - 2)))
        ctx.save_for_backward(x, weight, bias, mean, var, eps)

    @staticmethod
    def backward(ctx, ops, grad_output):
        x, weight, bias, mean, var, eps = ctx.saved_tensors

        var_eps = ops.add(var, eps)
        inv_std = ops.div(ops.scalar_from_float(1.0), ops.sqrt(var_eps))
        x_centered = ops.sub(x, mean)
        x_hat = ops.mul(x_centered, inv_std)

        grad_x = grad_weight = grad_bias = None
        if bias is not None:
            grad_bias = ops.sum(grad_output, dim=ctx.red_dims, keepdim=False)

        if weight is not None:
            grad_y_wrt_x_hat = ops.mul(grad_output, weight.reshape(1, -1, *([1] * (x.dim() - 2))))
            grad_weight = ops.sum(ops.mul(grad_output, x_hat), dim=ctx.red_dims, keepdim=False)
        else:
            grad_y_wrt_x_hat = grad_output

        if ctx.training:
            N = 1
            for dim in ctx.red_dims:
                N *= x.shape[dim]
            n_elems = ops.scalar_from_float(N)

            var_eps = ops.add(var, eps)
            inv_std_cubed = ops.mul(inv_std, ops.mul(inv_std, inv_std))
            neg_half = ops.scalar_from_float(-0.5)

            grad_var = ops.sum(ops.mul(
                ops.mul(grad_y_wrt_x_hat, x_centered),
                ops.mul(neg_half, inv_std_cubed)),
                dim=ctx.red_dims, keepdim=True)

            neg_inv_std = ops.neg(inv_std)
            grad_mean_term1 = ops.sum(ops.mul(grad_y_wrt_x_hat, neg_inv_std),
                                      dim=ctx.red_dims, keepdim=True)

            neg_two = ops.scalar_from_float(-2.0)
            neg_two_over_N = ops.div(neg_two, n_elems)
            sum_x_centered = ops.sum(x_centered, dim=ctx.red_dims, keepdim=True)
            grad_mean_term2 = ops.mul(ops.mul(grad_var, neg_two_over_N), sum_x_centered)

            grad_mean = ops.add(grad_mean_term1, grad_mean_term2)

            two = ops.scalar_from_float(2.0)
            two_over_N = ops.div(two, n_elems)
            one_over_N = ops.div(ops.scalar_from_float(1.0), n_elems)

            grad_x_term1 = ops.mul(grad_y_wrt_x_hat, inv_std)
            grad_x_term2 = ops.mul(ops.mul(grad_var, two_over_N), x_centered)
            grad_x_term3 = ops.mul(grad_mean, one_over_N)
            grad_x = ops.add(grad_x_term1, ops.add(grad_x_term2, grad_x_term3))

        else:
            grad_x = ops.mul(grad_y_wrt_x_hat, inv_std)

        return grad_x, None, None, grad_weight, grad_bias, None, None, None

@register_base_op("layer_norm")
def dt_layer_norm(ops, x, eps, normalized_shape, weight=None, bias=None):
    reduce_dims = tuple(range(x.dim() - len(normalized_shape), x.dim()))
    mean = ops.mean(x, dim=reduce_dims, keepdim=True)
    var = ops.var(x, ops.scalar_from_float(0.0), dim=reduce_dims, keepdim=True)

    var_eps = ops.add(var, eps)
    inv_std = ops.div(ops.scalar_from_float(1.0), ops.sqrt(var_eps))
    x_hat = ops.mul(ops.sub(x, mean), inv_std)

    if weight is not None:
        shape = [1] * (x.dim() - len(normalized_shape)) + list(normalized_shape)
        x_hat = ops.mul(x_hat, weight.view(*shape))
    if bias is not None:
        shape = [1] * (x.dim() - len(normalized_shape)) + list(normalized_shape)
        x_hat = ops.add(x_hat, bias.view(*shape))

    return x_hat

class DTLayerNormFunction(DTFunction):

    @staticmethod
    def forward(ops, x, eps, normalized_shape, weight=None, bias=None):
        return ops.layer_norm(x, eps, normalized_shape, weight, bias)

    @staticmethod
    def setup_context(ctx, ops, inputs, output):
        x, eps, normalized_shape, weight, bias = inputs
        ctx.red_dims = tuple(range(x.dim() - len(normalized_shape), x.dim()))
        mean = ops.mean(x, dim=ctx.red_dims, keepdim=True)
        var = ops.var(x, ops.scalar_from_float(0.0), dim=ctx.red_dims, keepdim=True)
        ctx.save_for_backward(x, weight, bias, mean, var, eps)

    @staticmethod
    def backward(ctx, ops, grad_output):
        x, weight, bias, mean, var, eps = ctx.saved_tensors
        red_dims = ctx.red_dims
        var_eps = ops.add(var, eps)
        inv_std = ops.div(ops.scalar_from_float(1.0), ops.sqrt(var_eps))
        x_centered = ops.sub(x, mean)
        x_hat = ops.mul(x_centered, inv_std)

        param_red_dims = tuple(i for i in range(x.dim()) if i not in red_dims)
        grad_bias = None
        if bias is not None:
            grad_bias = ops.sum(grad_output, dim=param_red_dims, keepdim=False)

        grad_weight = None
        if weight is not None:
            w_view = [1] * x.dim()
            for d in red_dims:
                w_view[d] = x.size(d)
            weight_b = weight.view(*w_view)
            grad_y_wrt_x_hat = ops.mul(grad_output, weight_b)
            grad_weight = ops.sum(
                ops.mul(grad_output, x_hat),
                dim=param_red_dims, keepdim=False
            )
        else:
            grad_y_wrt_x_hat = grad_output

        N = 1
        for d in red_dims:
            N *= x.shape[d]
        n_elems = ops.scalar_from_float(N)

        sum_grad = ops.sum(grad_y_wrt_x_hat, dim=red_dims, keepdim=True)
        sum_grad_xhat = ops.sum(
            ops.mul(grad_y_wrt_x_hat, x_hat),
            dim=red_dims, keepdim=True
        )

        Ng = ops.mul(n_elems, grad_y_wrt_x_hat)
        term_inner = ops.sub(
            ops.sub(Ng, sum_grad),
            ops.mul(x_hat, sum_grad_xhat)
        )
        inv_N = ops.div(ops.scalar_from_float(1.0), n_elems)
        grad_x = ops.mul(ops.mul(inv_std, inv_N), term_inner)

        return grad_x, None, None, grad_weight, grad_bias
